/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "dixie.h"
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/syslog.h>
#include <sys/wait.h>
#include <signal.h>

#ifndef FD_SET
#define NFDBITS         32
#define FD_SETSIZE      32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

#define DAPUSER			"c=US@o=University of Michigan@ou=Miscellaneous Servers@cn=Finger"
#define DEFAULT_BASE		"c=US@o=University of Michigan@ou=People"
#define DEFAULT_DIXIEHOST	"localhost"
#define DEFAULT_PORT		79
#define DEFAULT_SIZELIMIT	50
#define FINGER			"/usr/ucb/finger"

int	debug;
int	dosyslog = 1;

char	*dixiehost = DEFAULT_DIXIEHOST;
char	*base = DEFAULT_BASE;
int	strip = 3;
int	searchaliases = 1;

static usage( name )
char	*name;
{
	fprintf( stderr, "usage: %s [-d debuglevel] [-p port] [-l] [-x dixiehost] [-b searchbase] [-a] [-s stripcount] [-z sizelimit\r\n" );
	exit( 1 );
}

main (argc, argv)
int	argc;
char	**argv;
{
	int			s, ns, rc;
	int			port = -1;
	int			tblsize;
	int			i, pid;
	char			*myname;
	unsigned long		mypeer = -1;
	fd_set			readfds;
	struct hostent		*hp;
	struct timeval		timeout;
	struct sockaddr_in	from;
	int			fromlen;
	int			wait4child();
	extern char		*optarg;

	dxi_sizelimit = DEFAULT_SIZELIMIT;
	while ( (i = getopt( argc, argv, "ab:d:lp:s:x:" )) != EOF ) {
		switch( i ) {
		case 'a':	/* do not deref aliases when searching */
			dxi_options |= DXI_OPT_DONTSEARCHALIASES;
			break;

		case 'b':	/* search base */
			base = strdup( optarg );
			break;

		case 'd':	/* turn on debugging */
			debug = atoi( optarg );
			break;

		case 'l':	/* do syslogging */
			dosyslog = 1;
			break;

		case 's':	/* number of components to strip off dns */
			strip = atoi( optarg );
			break;

		case 'x':	/* specify dixie host */
			dixiehost = strdup( optarg );
			break;

		case 'z':	/* size limit */
			dxi_sizelimit = atoi( optarg );
			break;

		default:
			usage( argv[0] );
		}
	}
	if ( argc > 1 && *argv[1] != '-' ) {
		mypeer = (unsigned long) strtol( argv[1], NULL, 16 );
	}
	printf( "X.500 Finger Service...\r\n" );
	fflush( stdout );

	if ( (myname = strrchr( argv[0], '/' )) == NULL )
		myname = strdup( argv[0] );
	else
		myname = strdup( myname + 1 );

	if ( dosyslog && openlog( myname, LOG_PID | LOG_NOWAIT, LOG_LOCAL4 )
	    < 0 ) {
		perror("openlog");
	}

	if ( dosyslog && mypeer != -1 ) {
		struct in_addr	addr;

		hp = gethostbyaddr( (char *) &mypeer, sizeof(mypeer), AF_INET );
		addr.s_addr = mypeer;
		syslog( LOG_INFO, "connection from %s (%s)", (hp == NULL) ?
		    "unknown" : hp->h_name, inet_ntoa( addr ) );
	}

	do_query();
}

static char	dixie_buffer[MAX_PKT_SIZE];

static do_query()
{
	char		buf[256];
	int		len, i;
	int		rc, tblsize;
	struct timeval	timeout;
	fd_set		readfds;

	if ( dxi_init( dixiehost, 0, 0 ) == -1 ) {
		perror( "dxi_init" );
		exit( 1 );
	}

	if ( dxi_bind( DAPUSER, NULL ) == -1 ) {
		dxi_perror( "dxi_bind" );
		exit( 1 );
	}
	dxi_not61 = 1;

	tblsize = getdtablesize();
	timeout.tv_sec = 60;
	timeout.tv_usec = 0;
	FD_ZERO( &readfds );
	FD_SET( fileno( stdin ), &readfds );

	if ( (rc = select( tblsize, &readfds, 0, 0, &timeout )) <= 0 ) {
		if ( rc < 0 )
			perror( "select" );
		else
			fprintf( stderr, "connection timed out on input\r\n" );
		exit( 1 );
	}

	if ( fgets( buf, sizeof(buf), stdin ) == NULL )
		exit( 1 );

	len = strlen( buf );
	if ( debug ) {
		fprintf( stderr, "got %d bytes\r\n", len );
		dxi_bprint( buf, len );
	}

	/* strip off \r \n */
	if ( buf[len - 1] == '\n' ) {
		buf[len - 1] = '\0';
		len--;
	}
	if ( buf[len - 1] == '\r' ) {
		buf[len - 1] = '\0';
		len--;
	}

	/* turn .'s and _'s into spaces */
	for ( i = 0; i < len; i++ ) {
		if ( buf[i] == '.' || buf[i] == '_' || buf[i] == '*' )
			buf[i] = ' ';
	}

	if ( len == 0 ) {
		printf( "No campus-wide login information available.  Info for this machine only:\r\n" );
		fflush( stdout );
		execl( FINGER, FINGER, NULL );
	} else {
		do_search( buf );
	}
}

static do_search( buf )
char	*buf;
{
	char	filter[1024];
	char	*dn, *e, *rdn;
	char	**title;
	int	matches, spaces, i, approximate = 0;
	static char	*attrs[] = { "cn", "title", 0 };

	spaces = 0;
	for ( i = 0; buf[i] != '\0'; i++ ) {
		if ( isspace( buf[i] ) ) {
			spaces = 1;
			break;
		}
	}

	if ( spaces )
		sprintf( filter, "cn=%s|sn=%s", buf, buf, buf );
	else
		sprintf( filter, "sn=%s|uid=%s", buf, buf, buf );

	if ( dxi_search( base, DXI_SRCH_SUBTREE, filter, attrs, dixie_buffer )
	    == -1 ) {
		dxi_perror( "dxi_search" );
		exit( 1 );
	}

	matches = dxi_matches( dixie_buffer );
	if ( matches == 0 && (dxi_errno == DXI_NOTFOUND
	    || dxi_errno == DXI_SUCCESS) ) {
		if ( spaces )
			sprintf( filter, "cn~=%s|sn~=%s", buf, buf, buf );
		else
			sprintf( filter, "sn~=%s", buf, buf, buf );

		if ( dxi_search( base, DXI_SRCH_SUBTREE, filter, attrs,
		    dixie_buffer ) == -1 ) {
			dxi_perror( "dxi_search" );
			exit( 1 );
		}
		approximate = 1;
		matches = dxi_matches( dixie_buffer );
	}

	if ( dxi_errno == DXI_SIZELIMIT || dxi_errno == DXI_TIMELIMIT )
		printf( "(Partial results only - a limit was exceeded)\r\n" );

	if ( matches == 1 ) {
		if ( approximate )
			printf( "One approximate match found for '%s':\r\n",
			    buf );
		e = dxi_firstentry( dixie_buffer );
		do_read( dxi_getdn( e ) );
		free( dn );
	} else if ( matches > 1 ) {
		if ( approximate )
			printf( "%d approximate matches for '%s':\r\n",
			    matches, buf );
		else
			printf( "%d matches for '%s':\r\n", matches, buf );

		for ( e = dxi_firstentry( dixie_buffer ); e != NULL;
		    e = dxi_nextentry() ) {
			dn = dxi_getdn( e );
			rdn = dxi_lastdncomp( dn );
			if ( stricmp( rdn, buf ) == 0 ) {
				char	**cn;
				char	*s;
				int	i, last;

				cn = dxi_getvalues( e, "cn" );
				for ( i = 0; cn[i] != NULL; i++ ) {
					last = strlen( cn[i] ) - 1;
					if ( isdigit( cn[i][last] ) ) {
						rdn = strdup( cn[i] );
						break;
					}
				}
			}
					
			title = dxi_getvalues( e, "title" );

			printf( "  %-20s    %s\r\n", rdn,
			    title ? title[0] : "" );

			dxi_valuefree( title );
			free( dn );
		}
	} else if ( matches == 0 ) {
		printf( "Search failed to find anything.  Currently, this service contains information\r\n" );
		printf( "only on University of Michigan faculty, staff and students.  Other likely\r\n" );
		printf( "places to finger are:\r\n" );
		printf( "    um.cc.umich.edu\r\n" );
		printf( "    caen.engin.umich.edu\r\n" );
		printf( "    med.umich.edu\r\n" );
	} else {
		fprintf( stderr, "error return from dxi_matches\r\n" );
		exit( 1 );
	}
}

static print_attr( fp, label, attr, e, multiline )
FILE	*fp;
char	*label;
char	*attr;
char	*e;
int	multiline;
{
	char	**val;
	int	i, gotone = 0;

	if ( (val = dxi_getvalues( e, attr )) == NULL ) {
		if ( strcmp( attr, "mail" ) == 0 ) {
			fprintf( fp, "%-19s\r\n", label );
			fprintf( fp, "                    None registered in this service.  Try fingering at:\r\n" );
			fprintf( fp, "                    um.cc.umich.edu, caen.engin.umich.edu, med.umich.edu.\r\n" );
		}
		return;
	}

	fprintf( fp, "%-19s\r\n", label );
	for ( i = 0; val[i] != NULL; i++ ) {
		if ( multiline ) {
			char	*s, *p;

			if ( gotone )
				fprintf( fp, "%-19s\r\n", label );
			p = s = val[i];
			while ( s = strchr( s, '$' ) ) {
				*s++ = '\0';
				while ( isspace( *s ) )
					s++;
				fprintf( fp, "                    %s\r\n", p );
				p = s;
			}
			fprintf( fp, "                    %s\r\n", p );
			gotone = 1;
		} else {
			fprintf( fp, "                    %s\r\n", val[i] );
		}
	}
	dxi_valuefree( val );
}

static do_read( buf )
char	*buf;
{
	char		*dn, *e, *ufn;
	int		i;
	char		*s;
	char		**cn, **paddr, **tnumber, **uid, **title, **mail;
	static char	*attrs[] = { "cn", "postalAddress", "uid", "title",
				     "mail", "telephoneNumber",
				     "facsimileTelephoneNumber", 0 };

	if ( dxi_read( buf, attrs, dixie_buffer ) == -1 ) {
		dxi_perror( "dxi_read" );
		exit( 1 );
	}

	e = dxi_getreadentry( dixie_buffer );
	dn = dxi_getdn( e );
	ufn = dxi_dn2ufn( dn );
	for ( i = 0; i < strip; i++ ) {
		if ( (s = strrchr( ufn, ',' )) == NULL )
			break;
		*s = '\0';
	}
	fprintf( stdout, "\"%s\"\r\n", ufn );
	print_attr( stdout, "  Also known as:", "cn", e, 0 );
	print_attr( stdout, "  E-mail address:", "mail", e, 0 );
	print_attr( stdout, "  Fax number:", "facsimileTelephoneNumber", e, 0 );
	print_attr( stdout, "  Business phone:", "telephoneNumber", e, 0 );
	print_attr( stdout, "  Business address:", "postalAddress", e, 1 );
	print_attr( stdout, "  Title:", "title", e, 0 );
	print_attr( stdout, "  Uniqname:", "uid", e, 0 );

	free( dn );
}
